package com.fossgalaxy.games.tbs.ai.strategicmcts;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.ai.Controller;
import com.fossgalaxy.games.tbs.ai.rules.ProductionRuleAgent;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.games.tbs.parameters.EntityType;
import com.fossgalaxy.object.annotations.ObjectDef;
import rts.ai.evaluation.EvaluationFunction;

import java.util.*;
import java.util.stream.Collectors;

/**
 * ProductionRuleAgent[][] split first by "," and secondly by "/"
 * <p>
 * So separate sets of strategies for an EntityType with "," and individual strategies with "/"
 */
public class StrategicMCTS implements Controller {

    protected final Map<EntityType, List<ProductionRuleAgent>> rules;
    protected final int macroLength;
    protected int playerID;
    protected final EvaluationFunction eval;

    @ObjectDef("StrategicMCTS")
    public StrategicMCTS(EvaluationFunction eval,  int macroLength, EntityType[] types, ProductionRuleAgent[][] rulesForEntity) {
        assert types.length == rulesForEntity.length;
        rules = new HashMap<>();

        for (int i = 0; i < types.length; i++) {
            rules.put(types[i], Arrays.asList(rulesForEntity[i]));
        }

        this.macroLength = macroLength;
        this.eval = eval;
    }

    @Override
    public Map<UUID, Order> doTurn(int playerID, GameState state) {
        this.playerID = playerID;
        // TODO Move this to the node - every node needs to do this before their moves. Or before passing the state down to the child
        Collection<Entity> entitiesNeedingOrders = state.getOwnedEntities(playerID);
        Map<UUID, Order> automatedActions = doAutomatic(entitiesNeedingOrders, state);
        List<Entity> mctsEntities = entitiesNeedingOrders.stream().filter(x -> automatedActions.containsKey(x.getID())).collect(Collectors.toList());

        return null;
    }

    /**
     * Scan and find entities that can have automatic actions applied without bothering the MCTS.
     *
     * @param entitiesNeedingOrders The entities in the game that are ours
     * @return Orders generated from this
     */
    private Map<UUID, Order> doAutomatic(Collection<Entity> entitiesNeedingOrders, GameState state) {
        return entitiesNeedingOrders.stream()
                .filter(x -> x.getType().getAvailableActions().size() == 1)
                .filter(x -> x.getType().getAvailableActions().get(0).canAutomate())
                .collect(Collectors.toMap(Entity::getID, x -> x.getType().getAvailableActions().get(0).generateOrder(x.getPos(), state)));
    }
}

